
/* 
 * Process GRASS format LULC and DEM data to fix up holes (zeroes) in the 
 * data, and assign sensible color values to the various terrain features. 
 *
 * Program is called with input variables 
 *   a.out <DEM data input file> <LULC data input file> <output file>
 * and reads in LULC color map from file LULC.colors
 * 
 * Program outputs
 *   1) where the LULC and/or DEM values are 0 (to stdout)  
 *   2) A ppm file showing only these values
 *   3) A ppm file of the whole scene
 *   3) A data file of a grid of elevations and colors 
 *
*/


#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>

#define NumCategories 100
#define NumColors     3



main(int argc, char *argv[])
{
   FILE *InputDEM;
   FILE *InputLULC;

   FILE *ColorMap;
   unsigned char color[NumCategories][NumColors];  /* color map for LULC categories */
   int   OutputFile;

   int   BogusDataFile;
   char  Header[256];
   char  filename[256];
   char  line[1024];
   long  North = 0;
   long  South = 0;
   long  East = 0;
   long  West = 0;
   long  M = 0;
   long  N = 0;

   int  *DEM_Buffer;
   int  *Bogus_DEM_Buffer;
   unsigned char *LULC_Buffer;
   unsigned char *Bogus_LULC_Buffer;
   long  Buffer_Size;
   int   i;
   int   j;
   int   Index;
   int category;
   int red, green, blue;

   long  DEM;
   long  LULC;
   
   if (argc < 3)
   {
      printf("usage: <grass DEM datafile> <grass LULC datafile> <PPM \
datafile>\n");
      exit(1);
   }

   /* Open input file of DEM data */
   if ((InputDEM = fopen(argv[1], "r")) == NULL)
   {
      printf("Error -- cannot open %s\n", argv[1]);
      exit(1);
   }

   fscanf(InputDEM, "north: %ld%*c", &North);
   fscanf(InputDEM, "south: %ld%*c", &South);
   fscanf(InputDEM, "east: %ld%*c", &East);
   fscanf(InputDEM, "west: %ld%*c", &West);
   fscanf(InputDEM, "rows: %ld%*c", &M);
   fscanf(InputDEM, "cols: %ld%*c", &N);

   /* Open input file of LULC data */
   if ((InputLULC = fopen(argv[2], "r")) == NULL)
   {
      printf("Error -- cannot open %s\n", argv[2]);
      exit(1);
   }

   fscanf(InputLULC, "north: %*ld%*c");
   fscanf(InputLULC, "south: %*ld%*c");
   fscanf(InputLULC, "east: %*ld%*c");
   fscanf(InputLULC, "west: %*ld%*c");
   fscanf(InputLULC, "rows: %*ld%*c");
   fscanf(InputLULC, "cols: %*ld%*c");

   printf("(North, South, east, West) = (%ld, %ld, %ld, %ld)\n",
          North, South, East, West);
   printf("(row, col) = (%ld, %ld)\n", M, N);
   
   Buffer_Size = 3 * N * sizeof(unsigned char);

#if 0
   printf("Buffer_Size = %d\n", Buffer_Size);
#endif

   if ((DEM_Buffer = (int *) malloc(N * sizeof(int))) == NULL)
   {
      printf("Out of memory!\n");
      exit(1);
   }

   if ((LULC_Buffer = (unsigned char *) malloc(Buffer_Size)) == NULL)
   {
      printf("Out of memory!\n");
      exit(1);
   }

   if ((Bogus_DEM_Buffer = (int *) malloc(N * sizeof(int))) == NULL)
   {
      printf("Out of memory!\n");
      exit(1);
   }

   if ((Bogus_LULC_Buffer = (unsigned char *) malloc(Buffer_Size)) == NULL)
   {
      printf("Out of memory!\n");
      exit(1);
   }

   for (i = 0; i < NumCategories; i++)
      for (j = 0; j < NumColors; j++)
         color[i][j] = (unsigned char) 0;

   /* Read in the LULC color map */
   if ((ColorMap = fopen("LULC.color.map", "r")) == NULL)
   {
      printf("Error -- cannot open file LULC.color\n");
      exit(1);
   }

   printf("reading in the LULC color map ....\n");
   fflush(stdout);

   while (fgets(line, 1024, ColorMap) != NULL) 
   {
      printf("%s", line);
      fflush(stdout);

      /* screen for lines commented with a # */
      if (line[0] != '#')
      {
         sscanf(line, "%d %d %d %d", &category, &red, &green, &blue);
         color[category][0] = (unsigned char) red;
         color[category][1] = (unsigned char) green;
         color[category][2] = (unsigned char) blue;
         printf("%d %hd %hd %hd\n", category, 
                color[category][0], color[category][1], color[category][2]);
         fflush(stdout);
      }
   }
 
   fclose(ColorMap);

   /* Open output file for procesed data */
   sprintf(filename, "%s.ppm", argv[3]);
   if ((OutputFile = open(filename, O_WRONLY | O_CREAT | O_TRUNC,
                          S_IRUSR |  S_IWUSR)) == -1)
   {
      printf("Error -- cannot open %s\n", filename);
      exit(1);
   }

   sprintf(filename, "bugus-%s.ppm", argv[3]);
   if ((BogusDataFile = open(filename, O_WRONLY | O_CREAT | O_TRUNC,
                             S_IRUSR |  S_IWUSR)) == -1)
   {
      printf("Error -- cannot open %s\n", filename);
      exit(1);
   }

   sprintf(Header, "P6\n%d %d\n255\n\0", N,  M);

#if 0
   printf("Size of Header = %d length = %d\n", sizeof(Header), strlen(Header));
#endif

   write(OutputFile, Header, strlen(Header));
   write(BogusDataFile, Header, strlen(Header));

   printf("reading in the DEM and LULC data ....\n");
   fflush(stdout);

   /* Read in and process the DEM and LULC data */
   for (i = 0; i < M; i++)
   {
      for (j = 0, Index = 0; j < N; j++, Index += 3)
      {
         fscanf(InputDEM, "%ld%*c", &DEM);
         fscanf(InputLULC, "%ld%*c", &LULC);
         
         /* Currently this serves no purpose */
         DEM_Buffer[j] = DEM;
         Bogus_DEM_Buffer[j] = DEM;

         if (LULC > 0)
         {
            /* Assign colors using LULC color map */           
            LULC_Buffer[Index]     = color[LULC][0];
            LULC_Buffer[Index + 1] = color[LULC][1];
            LULC_Buffer[Index + 2] = color[LULC][2];
         
            if (DEM == 0)
            {
               /* Make pixels bright blue (water?) */           
               Bogus_LULC_Buffer[Index]     = (unsigned char) 0;
               Bogus_LULC_Buffer[Index + 1] = (unsigned char) 64;
               Bogus_LULC_Buffer[Index + 2] = (unsigned char) 64;

               printf("(i, j, DEM, LULC) = (%d, %d, %ld, %ld)\n", 
                        i, j, DEM, LULC);
            }
            else
            {
               /* Make pixels white */           
               Bogus_LULC_Buffer[Index]     = (unsigned char) 0;
               Bogus_LULC_Buffer[Index + 1] = (unsigned char) 0;
               Bogus_LULC_Buffer[Index + 2] = (unsigned char) 255;
            }
         }
         else
         {
            printf("(i, j, DEM, LULC) = (%d, %d, %ld, %ld)\n", i, j, DEM, LULC);
            
            /* Make pixels black */           
            LULC_Buffer[Index]     = (char) 0;
            LULC_Buffer[Index + 1] = (char) 0;
            LULC_Buffer[Index + 2] = (char) 0;
         
            if (DEM == 0)
            {
               /* Make pixels dark blue */           
               Bogus_LULC_Buffer[Index]     = (char) 0;
               Bogus_LULC_Buffer[Index + 1] = (char) 255;
               Bogus_LULC_Buffer[Index + 2] = (char) 0;
            }
            else
            {
               /* Make pixels black */           
               Bogus_LULC_Buffer[Index]     = (char) 0;
               Bogus_LULC_Buffer[Index + 1] = (char) 0;
               Bogus_LULC_Buffer[Index + 2] = (char) 0;
            }
         }
      }

      /* Print out row of LULC ppm file */
      write(OutputFile, LULC_Buffer, Buffer_Size);

      /* Print out row showing bogus LULC data to a ppm file */
      write(BogusDataFile, Bogus_LULC_Buffer, Buffer_Size);
   }

   close(OutputFile);
   close(BogusDataFile);

   fclose(InputDEM);
   fclose(InputLULC);
}



